#!/usr/bin/perl -w

use strict;

my( $switch );
my $opt_filename = "";
my $opt_tabsize = 4;	#default tab size
my $opt_tabnum = 2;		#default number of tabs
my $opt_commenttabnum;	#default opt_tabnum+2 (calculated after parameters are read)
my $opt_commentstring = ";"; #default is ";".. Can be changed to "//" etc.	
my $opt_debug = 0;		#default is debug = false
my $opt_notabs = 0;		#default is notabs = false
my $opt_addcolon = 0;	#default is force colon at end of labels = false
my $opt_removetrailingspaces = 0; #default is no removal of trailing spaces at end of lines

#Get command line parameters
while ($switch = shift @ARGV) {
	if ($switch eq "-t") {  #+ARG
		$opt_tabnum = shift or die("Option -t requires a parameter (number of tabs to code column).\n");
	} elsif ($switch eq "-s") { #+ARG
		$opt_tabsize = shift or die("Option -s requires a parameter (number of spaces that correspond to one tab).\n");
	} elsif ($switch eq "-c") {#+ARG
		$opt_commenttabnum = shift or die("Option -c requires a parameter (number of tabs to comment column).\n");
	} elsif ($switch eq "-m") { #+ARG
		$opt_commentstring = shift or die("Option -m requires a parameter (comment marker char used in infile).\n");
	} elsif ($switch eq "-n") {
		$opt_notabs = 1; #Use spaces instead of tab characters
	} elsif ($switch eq "-C") {
		$opt_addcolon = 1; #Force colon to be added at end of label
	} elsif ($switch eq "-x") {
		$opt_removetrailingspaces = 1; #Remove trailing spaces at end of lines
	} elsif ($switch eq "-h" or $switch eq "--help" or $switch eq "-?") {
		printhelp();
	} elsif ($switch eq "-d") {
		$opt_debug = 1; #Secret option
	} else {
		#Assume that a filename was encountered.
		last;
	}
}

#Some more validation of parameters
$opt_commenttabnum = $opt_tabnum+4 if (!$opt_commenttabnum); #Default setting of column to use for 
die ("Comment column must be greater than code column.") if ($opt_commenttabnum < $opt_tabnum);

#Did we get a file name?
my @infile;
if ($switch) {
	$opt_filename = $switch;

	#Open file
	open(DAT, $opt_filename) or die "Couldn't open file \"$opt_filename\"! : $!\n";
	@infile = <DAT>; #Slurp file
	close(DAT);
} else {
	#@infile = <STDIN>;
	#Slurping files from STDIN can be dangerous, so do the old fashioned line-by-line reading
	while(<STDIN>) {
    	push @infile, $_ ;
	}
}

indentfile(\@infile) or die("ERROR: Failed to indent $opt_filename for some unknown reason!\n");

#Quit program
print STDERR "[All done!]\n";
exit 0;

######################################################################################
#Print help
sub printhelp {
		print "\n";
		print "6510i v0.01 by FTC\n";
		print "\n";
		print "Indents a 6502/6510 assembler source file and outputs the result to stdout.\n";
		print "\n";
		print "WARNING: DO NOT OVERWRITE THE INFILE WITH THE OUTPUT OF THIS PROGRAM.\n";
		print "\n";
		print "USAGE: 6510i.pl [options [arguments]] INFILE\n";
		print "USAGE: 6510i.pl [options [arguments]] INFILE > OUTFILE\n";
		print "USAGE: cat INFILE | 6510i.pl [options [arguments]]\n";
		print "\n";
		print "\t-t N\tTab column for code (default 2).\n";
		print "\t-s N\tNumber of spaces corresponding to one tab (default 4).\n";
		print "\t-c N\tTab column for comments after code. Must be greater than tab column for code. (default codetabs+4).\n";
		print "\t-m STR\tSpecify comment marker (most use \";\" but KickAssembler uses \"//\").\n";
		print "\t-n\tNo tabs. Convert all tab characters to spaces.\n";
		print "\t-C\tForce addition of colon to labels.\n";
		print "\t-x\tRemove trailing spaces at end of lines.\n";
		print "\t-h\tPrint this help. Equivalent with --help and -?.\n";
		print "\n";
		exit 0;
}

######################################################################################
#Indent file
sub indentfile {
	
	my $infile = shift;
		
	#Build tab-string
	my $codecolumn = $opt_tabsize*$opt_tabnum;
	my $commentcolumn = $opt_tabsize*$opt_commenttabnum;
	my $onetab = "";
	my $onetab_spaces = " " x $opt_tabsize;
	my $tabstring = "";
	my $commenttabstring = "";
	if ($opt_notabs) {
	 	$onetab = $onetab_spaces;
	} else {
		$onetab = "\t";
	}
	
	#Readymade tab strings from the beginning of the line to the code column and to the comment column respectively
	$tabstring = $onetab x $opt_tabnum;
	$commenttabstring = $onetab x $opt_commenttabnum;
	
	#Loop through the lines
	my $emptylines = 1; #Start with the assumption that line -1 was empty
	foreach my $line (@infile) {

		my $debugstring = "";	#This string is filled with ";COMMENTS" placed at the end of the current line, for debugging
		my $outline = "";		#This is the resulting line of the indent "algorithm"
		chomp($line);			#Remove newline at end of line

		#Remove trailing spaces at end of line, if this option is enabled
		if ($opt_removetrailingspaces) {
			$line =~ s/\s*$//;
		}
		
		#Replace several empty lines in a row with only one empty line
		if ($line =~ m/^\s*$/) { #Empty line, or line with only spaces/tabs
			#Empty line found
			$emptylines++;
			if ($emptylines > 1) {
				next;
			} else {
				$emptylines = 1;
				#print "$line\n";
				print "\n";
				next;
			}
		}
		$emptylines = 0;

		#Convert all tabs on the line to spaces
		#xxx detta kan göras mer intelligent, så att t.ex. inte kommentarer som kommer efter kod förskjuts
		$line =~ s/\t/$onetab/g if ($opt_notabs);
		
		#Capture any comments
		my $firstpart;
		my $comment;
		my $commentpos;
		if ($line =~ s/^([^;]*)($opt_commentstring.*)$/$1/) { #Capture comment part of the line, if any
			$firstpart = $1;
			$comment = $2;
			#Estimate position of the comment
			$firstpart =~ s/\t/$onetab_spaces/g; #This is not perfect.. all tabs do not correspond to $tabsize number of spaces
			$commentpos = length($firstpart);
		}
		
		#Capture any labels
		my $label;
		if ($line =~ s/^([^\s]+:?|\s*[^\s]*:)(.*)/$2/) { #Capture two types of labels (beginning at col0 OR ending with ":")
			$label = $1;
			$label =~ s/^\s*//; #Remove leading spaces
			
			if ($opt_addcolon and $label !~ m/:$/) {
				$label .= ":"; #Add colon
			}
		}
		
		#Capture rest (code or expressions like "= $002+5/2" that are assigned to a label)
		my $code;
		if ($line =~ s/^\s*(.*)$/$1/) {
			$code = $1;
			$code =~ s/^\s*//; #Remove leading spaces of the "code" part
			$code =~ s/\s*$//; #Remove trailing spaces of the "code" part
		}
		
		#===========================
		#Build new output string
		
		my $linelength = 0;
		my $commentflushed = 0;

		#Flush col0 stuff (labels or comments at beginning of line)
		if ($label) {
			$outline .= "$label";
			$linelength += length($label);
		} elsif ($comment and !$commentflushed and $commentpos < $opt_tabsize) {
			$outline .= "$comment";
			$linelength += length($comment);
			$commentflushed = 1;
		}

		#Pad with tabs/spaces to the code column (But only if there is content to come
		if (($code or ($comment and !$commentflushed)) and $linelength < $codecolumn) {
			#Fill up with tabs/spaces between label and code/comments
			if ($opt_notabs) {
				$outline .= " " x ($codecolumn - $linelength);
				$linelength += ($codecolumn - $linelength);
			} else {
				my $padtabs = quotient(($codecolumn - $linelength +$opt_tabsize-1),$opt_tabsize);
				$outline .= $onetab x $padtabs;
				$linelength += $opt_tabsize*$padtabs;
			}
		} else {
			$outline .= " "; #Code column already passed. Out of alignment. Print only one space character
			$linelength++;
		}

		#Flush code column stuff
		if ($code) {
			$outline .= "$code";
			$linelength += length($code);
		} elsif ($comment and !$commentflushed and $commentpos < $opt_commenttabnum*$opt_tabsize) {
			$outline .= "$comment";
			$linelength += length($comment);
			$commentflushed = 1;		
		}

		#Flush any remaning comment stuff to the comment column
		if ($comment and !$commentflushed) {
						
			#Pad with tabs/spaces to the code column
			if ($linelength < $commentcolumn) {
				#Fill up with tabs/spaces between codecolumn and comment column
				if ($opt_notabs) {
					$outline .= " " x ($commentcolumn - $linelength);
					$linelength += ($commentcolumn - $linelength);
				} else {
					my $padtabs = quotient(($commentcolumn - $linelength +$opt_tabsize-1),$opt_tabsize);
					$outline .= $onetab x $padtabs;
					$linelength += $opt_tabsize*$padtabs;
				}
			} else {
				#What we do if comment column already passed and we're out of alignment
				$outline .= $onetab; #Add tab
				$linelength += $opt_tabsize;
			}

			#Flush the comment
			$outline .= "$comment";
		}
		
		#Always end line with a new line character
		print "$outline$debugstring\n";
	}
	
	return 1; #Success... hopefully
}

##############
# Integer division (Instead of using "use Integer;" in a local scope..)
#...because the exact division implementations seems to differ between different platform. Huu!
#
# parameters are $n then $d
# 
sub quotient {
	my $n = shift;
	my $d = shift;
	my $r = $n;
	my $q = 0;
	while ($r >= $d) { 
		$r = $r - $d;
		$q = $q + 1;
	}
	return $q;
}

